-------------------------------------------------------------------------------
-- LoreBooks v1.5.5
-------------------------------------------------------------------------------
--
-- Copyright (c) 2014, 2015 Ales Machat (Garkin)
--
-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.
--
-------------------------------------------------------------------------------
--
-- DISCLAIMER:
--
-- This Add-on is not created by, affiliated with or sponsored by ZeniMax
-- Media Inc. or its affiliates. The Elder Scrolls® and related logos are
-- registered trademarks or trademarks of ZeniMax Media Inc. in the United
-- States and/or other countries. All rights reserved.
--
-- You can read the full terms at:
-- https://account.elderscrollsonline.com/add-on-terms
--
-------------------------------------------------------------------------------

--Libraries--------------------------------------------------------------------
local LAM = LibStub("LibAddonMenu-2.0")
local LMP = LibStub("LibMapPins-1.0")

--Local constants -------------------------------------------------------------
local ADDON_VERSION = "1.5.5"
local PINS_UNKNOWN = "LBooksMapPin_unknown"
local PINS_COLLECTED = "LBooksMapPin_collected"
local PINS_COMPASS = "LBooksCompassPin_unknown"

--Local variables -------------------------------------------------------------
local updatePins = {}
local updating = false
local savedVariables          --user settings
local svData                  --collected books
local newBooks = {}           --books which are not listed in LoreBooksData.lua
local defaults = {            --default settings for saved variables
   compassMaxDistance = 0.04,
   pinTexture = {
      type = 1,
      size = 26,
      level = 40,
   },
   filters = {
      [PINS_COMPASS] = true,
      [PINS_UNKNOWN] = true,
      [PINS_COLLECTED] = false,
   },
}
local dataDefaults = {
   lorebooks = {},
}
local alliance = GetUnitAlliance("player")

-- Local functions ------------------------------------------------------------
--prints message to chat
local function MyPrint(...)
   CHAT_SYSTEM:AddMessage(...)
end

--save data to the table
local function SetLocalData(dataTable, zone, subzone, localData)
   dataTable = dataTable or {}
   dataTable[zone] = dataTable[zone] or {}
   dataTable[zone][subzone] = dataTable[zone][subzone] or {}

   table.insert(dataTable[zone][subzone], localData)
end

--checks if normalized point is valid
local function InvalidPoint(x, y)
   return (x < 0 or x > 1 or y < 0 or y > 1)
end

local function GetEpsilon(zone, subzone)
   --Epsilon value should be bigger for small maps and smaller for big maps.
   -- As I have to determine epsilon even when it's not possible to check
   -- mapsize directly, I must find some kind of workaround...
   local epsilon = 0.007

   if subzone == "ava_whole" then                     --Cyrodiil zone map
      epsilon = 0.003
   elseif zone == "cyrodiil" then                     --all other subzones in Cyrodiil
      epsilon = 0.010
   elseif (subzone):find(zone) then                   --zones
      epsilon = 0.005
   elseif zone == "guildmaps" or zone == "main" then  --Guild Maps, Main quest maps
      epsilon = 0.012
   end

   return epsilon
end

local function UnknownCoords(dataTable, x, y, epsilon)
   if dataTable == nil or dataTable == {} then
      return true
   end

   epsilon = epsilon or 0.005

   for i, data in ipairs(dataTable) do
      if zo_floatsAreEqual(data[1], x, epsilon) and zo_floatsAreEqual(data[2], y, epsilon) then
         return false
      end
   end

   return true
end

-- Pins -----------------------------------------------------------------------
local pinTexturesList = {
   [1] = "Shalidor's Library icons",
   [2] = "Book icon set 1",
   [3] = "Book icon set 2",
   [4] = "Esohead's icons (Rushmik)",
}

local pinTextures = {
   --[index] = { known_book_texture, unknown_book_texture },
   [1] = { "EsoUI/Art/Icons/lore_book4_detail4_color5.dds", "EsoUI/Art/Icons/lore_book4_detail4_color5.dds" },
   [2] = { "LoreBooks/Icons/book1.dds", "LoreBooks/Icons/book1-invert.dds" },
   [3] = { "LoreBooks/Icons/book2.dds", "LoreBooks/Icons/book2-invert.dds" },
   [4] = { "LoreBooks/Icons/book3.dds", "LoreBooks/Icons/book3-invert.dds" },
}

local function GetPinTexture(self)
   local _, texture, known = GetLoreBookInfo(1, self.m_PinTag[3], self.m_PinTag[4])
   local textureType = savedVariables.pinTexture.type

   return (pinTexturesList[textureType] == "Shalidor's Library icons") and texture or pinTextures[textureType][known and 1 or 2]
end

local function IsPinGrayscale()
   return pinTexturesList[savedVariables.pinTexture.type] == "Shalidor's Library icons"
end

--tooltip creator
local pinTooltipCreator = {}
pinTooltipCreator.tooltip = 1 --TOOLTIP_MODE.INFORMATION
pinTooltipCreator.creator = function(pin)
   local pinTag = pin.m_PinTag
   local title, icon, known = GetLoreBookInfo(1, pinTag[3], pinTag[4])
   local collection = GetLoreCollectionInfo(1, pinTag[3])
   local moreinfo = {}

   if pinTag[6] ~= nil then
      table.insert(moreinfo, "[" .. GetString("LBOOKS_MOREINFO", pinTag[6]) .. "]")
   end
   if pinTag[5] ~= nil then
      table.insert(moreinfo, "[" .. zo_iconFormat(GetAllianceSymbolIcon(pinTag[5]), 24, 24) .. "]")
   end

   InformationTooltip:AddLine(zo_strformat(collection), "ZoFontGameOutline", ZO_SELECTED_TEXT:UnpackRGB())
   ZO_Tooltip_AddDivider(InformationTooltip)
   InformationTooltip:AddLine(zo_iconTextFormat(icon, 32, 32, title), "", ZO_HIGHLIGHT_TEXT:UnpackRGB())
   if known then
      table.insert(moreinfo, "[" .. GetString(LBOOKS_KNOWN) .. "]")
   end
   if moreinfo[1] then
      InformationTooltip:AddLine(zo_strjoin(" / ", moreinfo), "", ZO_TOOLTIP_DEFAULT_COLOR:UnpackRGB())
   end
end

--add callbacks
local function CreatePins()
   local zone, subzone = LMP:GetZoneAndSubzone()
   local lorebooks = LoreBooks_GetLocalData(zone, subzone)

   if lorebooks ~= nil then
      for _, pinData in ipairs(lorebooks) do
         local _, _, known = GetLoreBookInfo(1, pinData[3], pinData[4])
         if known and updatePins[PINS_COLLECTED] and LMP:IsEnabled(PINS_COLLECTED) then
            LMP:CreatePin(PINS_COLLECTED, pinData, pinData[1], pinData[2])
         elseif not known then
            if updatePins[PINS_UNKNOWN] and LMP:IsEnabled(PINS_UNKNOWN) then
               LMP:CreatePin(PINS_UNKNOWN, pinData, pinData[1], pinData[2])
            end
            if updatePins[PINS_COMPASS] and savedVariables.filters[PINS_COMPASS] then
               COMPASS_PINS.pinManager:CreatePin(PINS_COMPASS, pinData, pinData[1], pinData[2])
            end
         end
      end
   end

   updatePins = {}
   updating = false
end

local function QueueCreatePins(pinType)
   updatePins[pinType] = true

   if not updating then
      updating = true
      if IsPlayerActivated() then
         zo_callLater(CreatePins, 50)
      else
         EVENT_MANAGER:RegisterForEvent("LoreBooks_PinUpdate", EVENT_PLAYER_ACTIVATED,
            function(event)
               EVENT_MANAGER:UnregisterForEvent("LoreBooks_PinUpdate", event)
               CreatePins()
            end)
      end
   end
end

local function MapCallback_unknown()
   if not LMP:IsEnabled(PINS_UNKNOWN) or (GetMapType() > MAPTYPE_ZONE) then return end
   QueueCreatePins(PINS_UNKNOWN)
end

local function MapCallback_collected()
   if not LMP:IsEnabled(PINS_COLLECTED) or (GetMapType() > MAPTYPE_ZONE) then return end
   QueueCreatePins(PINS_COLLECTED)
end

local function CompassCallback()
   if not savedVariables.filters[PINS_COMPASS] or (GetMapType() > MAPTYPE_ZONE) then return end
   QueueCreatePins(PINS_COMPASS)
end

-- Slash commands -------------------------------------------------------------
local function ShowMyPosition()
   local x, y = GetMapPlayerPosition("player")

   local locX = ("%05.02f"):format(zo_round(x*10000)/100)
   local locY = ("%05.02f"):format(zo_round(y*10000)/100)

   MyPrint(zo_strformat("<<1>>: <<2>>\195\151<<3>> (<<4>>)", GetMapName(), locX, locY, LMP:GetZoneAndSubzone(true)))
end
SLASH_COMMANDS["/mypos"] = ShowMyPosition
SLASH_COMMANDS["/myposition"] = ShowMyPosition
SLASH_COMMANDS["/myloc"] = ShowMyPosition
SLASH_COMMANDS["/mylocation"] = ShowMyPosition

local function ShowReport(textArray)
   local text = zo_strjoin("\n", textArray)
   LoreBooks_GUIText:SetText(text)
   LoreBooks_GUI:SetHidden(false)
   SetGameCameraUIMode(true) --release mouse
end

local function BuildReport(data)
   local dataTable = data
   local list = {}

   for zone, zoneData in pairs(dataTable) do
      for subzone, subzoneData in pairs(zoneData) do
         table.insert(list, zone .. "/" .. subzone)
         for _, pinTag in ipairs(subzoneData) do
            local bookname = GetLoreBookInfo(1, pinTag[3], pinTag[4])
            local tag = "{ "
            for i = 1, 3 do tag = tag .. tostring(pinTag[i]) .. ", " end
            tag = tag .. tostring(pinTag[4])
            if (pinTag[5] ~= nil) or (pinTag[6] ~= nil) then
               tag = tag .. ", " .. tostring(pinTag[5])
               if (pinTag[6] ~= nil) then
                  tag = tag .. ", " .. tostring(pinTag[6])
               end
            end
            tag = tag .. " }, --" .. bookname
            table.insert(list, tag)
         end
      end
      table.insert(list, "")
   end

   return list
end
SLASH_COMMANDS["/lbreport"] = function() ShowReport(BuildReport(newBooks)) end
SLASH_COMMANDS["/lbcollected"] = function() ShowReport(BuildReport(svData.lorebooks)) end

local function LoreBooksCheck()
   local bookCheck = {}
   local unknownBooks = {}

   --collect unknown books
   local categoryId = 1 --Shalidor's Library
   local _, numCollections = GetLoreCategoryInfo(categoryId)
   for collectionId = 1, numCollections do
      local _, _, numKnownBooks, totalBooks, hidden = GetLoreCollectionInfo(categoryId, collectionId)
      if not hidden and numKnownBooks ~= totalBooks then
         unknownBooks[collectionId] = {}
         for bookId = 1, totalBooks do
            local title, icon, known = GetLoreBookInfo(categoryId, collectionId, bookId)
            if not known then
               unknownBooks[collectionId][bookId] = {}
            end
         end
      end
   end

   --fill locations for unknown books
   local lorebooksData = LoreBooks_GetAllData()
   for zone, zoneData in pairs(lorebooksData) do
      for subzone, subzoneData in pairs(zoneData) do
         for i, pin in ipairs(subzoneData) do
            if unknownBooks[pin[3]] ~= nil and unknownBooks[pin[3]][pin[4]] ~= nil then
               local locationInfo = zo_strformat("<<1>>/<<2>>: <<3>>,<<4>>", zone, subzone, ("%03d"):format(pin[1]*1000), ("%03d"):format(pin[2]*1000))
               table.insert(unknownBooks[pin[3]][pin[4]], locationInfo)
            end
         end
      end
   end

   --build report
   for collectionId, books in pairs(unknownBooks) do
      local collectionName = GetLoreCollectionInfo(categoryId, collectionId)
      table.insert(bookCheck, zo_strformat("|cFFA500<<1>>|r", collectionName))

      for bookId, locations in pairs(books) do
         local title, icon = GetLoreBookInfo(categoryId, collectionId, bookId)
         table.insert(bookCheck, zo_strformat("|cFFFF00<<1>>|r", title))

         if locations == {} then
            table.insert(bookCheck, "no known locations")
         else
            for i, locationInfo in ipairs(locations) do
               table.insert(bookCheck, locationInfo)
            end
         end
      end
      table.insert(bookCheck, "")
   end

   return bookCheck
end
SLASH_COMMANDS["/lbcheck"] = function() ShowReport(LoreBooksCheck()) end

-- Settings menu --------------------------------------------------------------
local function CreateSettingsMenu()
   local panelData = {
      type = "panel",
      name = GetString(LBOOKS_TITLE),
      displayName = ZO_HIGHLIGHT_TEXT:Colorize(GetString(LBOOKS_TITLE)),
      author = "Garkin",
      version = ADDON_VERSION,
      slashCommand = "/lorebooks",
      registerForRefresh = true,
      registerForDefaults = true,
   }
   LAM:RegisterAddonPanel("LoreBooks", panelData)

   local CreateIcons, unknownIcon, collectedIcon
   CreateIcons = function(panel)
      if panel == LoreBooks then
         unknownIcon = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[1], CT_TEXTURE)
         unknownIcon:SetAnchor(RIGHT, panel.controlsToRefresh[1].combobox, LEFT, -10, 0)
         unknownIcon:SetTexture(pinTextures[savedVariables.pinTexture.type][2])
         unknownIcon:SetDimensions(savedVariables.pinTexture.size, savedVariables.pinTexture.size)
         collectedIcon = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[1], CT_TEXTURE)
         collectedIcon:SetAnchor(RIGHT, unknownIcon, LEFT, -5, 0)
         collectedIcon:SetTexture(pinTextures[savedVariables.pinTexture.type][1])
         collectedIcon:SetDimensions(savedVariables.pinTexture.size, savedVariables.pinTexture.size)
         collectedIcon:SetDesaturation((pinTexturesList[savedVariables.pinTexture.type] == "Shalidor's Library icons") and 1 or 0)
         CALLBACK_MANAGER:UnregisterCallback("LAM-PanelControlsCreated", CreateIcons)
      end
   end
   CALLBACK_MANAGER:RegisterCallback("LAM-PanelControlsCreated", CreateIcons)

   local optionsTable = {
      {
         type = "dropdown",
         name = GetString(LBOOKS_PIN_TEXTURE),
         tooltip = GetString(LBOOKS_PIN_TEXTURE_DESC),
         choices = pinTexturesList,
         getFunc = function() return pinTexturesList[savedVariables.pinTexture.type] end,
         setFunc = function(selected)
               for index, name in ipairs(pinTexturesList) do
                  if name == selected then
                     savedVariables.pinTexture.type = index
                     unknownIcon:SetTexture(pinTextures[index][2])
                     collectedIcon:SetDesaturation((pinTexturesList[index] == "Shalidor's Library icons") and 1 or 0)
                     collectedIcon:SetTexture(pinTextures[index][1])
                     LMP:RefreshPins(PINS_UNKNOWN)
                     LMP:RefreshPins(PINS_COLLECTED)
                     COMPASS_PINS.pinLayouts[PINS_COMPASS].texture = pinTextures[index][2]
                     COMPASS_PINS:RefreshPins(PINS_COMPASS)
                     break
                  end
               end
            end,
         disabled = function() return not (savedVariables.filters[PINS_UNKNOWN] or savedVariables.filters[PINS_COLLECTED]) end,
         default = pinTexturesList[defaults.pinTexture.type],
      },
      {
         type = "slider",
         name = GetString(LBOOKS_PIN_SIZE),
         tooltip = GetString(LBOOKS_PIN_SIZE_DESC),
         min = 20,
         max = 70,
         step = 1,
         getFunc = function() return savedVariables.pinTexture.size end,
         setFunc = function(size)
               savedVariables.pinTexture.size = size
               unknownIcon:SetDimensions(size, size)
               collectedIcon:SetDimensions(size, size)
               LMP:SetLayoutKey(PINS_UNKNOWN, "size", size)
               LMP:SetLayoutKey(PINS_COLLECTED, "size", size)
               LMP:RefreshPins(PINS_UNKNOWN)
               LMP:RefreshPins(PINS_COLLECTED)
            end,
         disabled = function() return not (savedVariables.filters[PINS_UNKNOWN] or savedVariables.filters[PINS_COLLECTED]) end,
         default = defaults.pinTexture.size
      },
      {
         type = "slider",
         name = GetString(LBOOKS_PIN_LAYER),
         tooltip = GetString(LBOOKS_PIN_LAYER_DESC),
         min = 10,
         max = 200,
         step = 5,
         getFunc = function() return savedVariables.pinTexture.level end,
         setFunc = function(level)
               savedVariables.pinTexture.level = level
               LMP:SetLayoutKey(PINS_UNKNOWN, "level", level)
               LMP:SetLayoutKey(PINS_COLLECTED, "level", level)
               LMP:RefreshPins(PINS_UNKNOWN)
               LMP:RefreshPins(PINS_COLLECTED)
            end,
         disabled = function() return not (savedVariables.filters[PINS_UNKNOWN] or savedVariables.filters[PINS_COLLECTED]) end,
         default = defaults.pinTexture.level,
      },
      {
         type = "checkbox",
         name = GetString(LBOOKS_UNKNOWN),
         tooltip = GetString(LBOOKS_UNKNOWN_DESC),
         getFunc = function() return savedVariables.filters[PINS_UNKNOWN] end,
         setFunc = function(state)
               savedVariables.filters[PINS_UNKNOWN] = state
               LMP:SetEnabled(PINS_UNKNOWN, state)
            end,
         default = defaults.filters[PINS_UNKNOWN],
      },
      {
         type = "checkbox",
         name = GetString(LBOOKS_COLLECTED),
         tooltip = GetString(LBOOKS_COLLECTED_DESC),
         getFunc = function() return savedVariables.filters[PINS_COLLECTED] end,
         setFunc = function(state)
               savedVariables.filters[PINS_COLLECTED] = state
               LMP:SetEnabled(PINS_COLLECTED, state)
            end,
         default = defaults.filters[PINS_COLLECTED]
      },
      {
         type = "checkbox",
         name = GetString(LBOOKS_COMPASS_UNKNOWN),
         tooltip = GetString(LBOOKS_COMPASS_UNKNOWN_DESC),
         getFunc = function() return savedVariables.filters[PINS_COMPASS] end,
         setFunc = function(state)
               savedVariables.filters[PINS_COMPASS] = state
               COMPASS_PINS:RefreshPins(PINS_COMPASS)
            end,
         default = defaults.filters[PINS_COMPASS],
      },
      {
         type = "slider",
         name = GetString(LBOOKS_COMPASS_DIST),
         tooltip = GetString(LBOOKS_COMPASS_DIST_DESC),
         min = 1,
         max = 100,
         step = 1,
         getFunc = function() return savedVariables.compassMaxDistance * 1000 end,
         setFunc = function(maxDistance)
               savedVariables.compassMaxDistance = maxDistance / 1000
               COMPASS_PINS.pinLayouts[PINS_COMPASS].maxDistance = maxDistance / 1000
               COMPASS_PINS:RefreshPins(PINS_COMPASS)
            end,
         disabled = function() return not savedVariables.filters[PINS_COMPASS] end,
         default = defaults.compassMaxDistance * 1000,
      },
   }
   LAM:RegisterOptionControls("LoreBooks", optionsTable)
end

-- Event handlers -------------------------------------------------------------
local function HandleBookLearned(x, y, collectionIndex, bookIndex)
   --get map information
   local zone, subzone = LMP:GetZoneAndSubzone()
   local lorebooks = LoreBooks_GetLocalData(zone, subzone)

   --build pin tag
   local locX = zo_round(x*1000) / 1000
   local locY = zo_round(y*1000) / 1000

   local pinData = {locX, locY, collectionIndex, bookIndex}
   if (collectionIndex > 6 and collectionIndex < 11) or (collectionIndex > 11 and collectionIndex < 20) then
      table.insert(pinData, alliance)
   end

   --if location is not known, save data to newBooks and lorebooksData
   --(will be stored just till logout/reload UI)
   local epsilon = GetEpsilon(zone, subzone)
   if UnknownCoords(lorebooks, x, y, epsilon) then
      SetLocalData(newBooks, zone, subzone, pinData)
      LoreBooks_SetLocalData(zone, subzone, pinData)
   end
   --save to saved variables for later use
   if UnknownCoords(svData.lorebooks, x, y, epsilon) then
      SetLocalData(svData.lorebooks, zone, subzone, pinData)
   end

   LMP:RefreshPins(PINS_UNKNOWN)
   LMP:RefreshPins(PINS_COLLECTED)
   COMPASS_PINS:RefreshPins(PINS_COMPASS)
end

local function OnBookLearned(eventCode, categoryIndex, collectionIndex, bookIndex)
   --categoryIndex = 1 for lore library
   --categoryIndex = 2 for all other books
   if categoryIndex == 1 then
      --get player position
      local mapChanged = SetMapToPlayerLocation() == SET_MAP_RESULT_MAP_CHANGED
      local x, y = GetMapPlayerPosition("player")

      if not InvalidPoint(x, y) then   --for unknown reason it sometime fails
         HandleBookLearned(x, y, collectionIndex, bookIndex)
      else                             --so try again after 250ms (conflict with another addon?)
         zo_callLater(function()
               if SetMapToPlayerLocation() == SET_MAP_RESULT_MAP_CHANGED then
                  mapChanged = true
               end
               x, y = GetMapPlayerPosition("player")
               if not InvalidPoint(x, y) then
                  HandleBookLearned(x, y, collectionIndex, bookIndex)
               else
                  LMP:RefreshPins(PINS_UNKNOWN)
                  LMP:RefreshPins(PINS_COLLECTED)
                  COMPASS_PINS:RefreshPins(PINS_COMPASS)
               end
               if mapChanged then
                  mapChanged = false
                  CALLBACK_MANAGER:FireCallbacks("OnWorldMapChanged")
               end
            end, 250)
      end
      if mapChanged then
         mapChanged = false
         CALLBACK_MANAGER:FireCallbacks("OnWorldMapChanged")
      end
   end
end

--updates lorebooksData from saved variables, if there is unknown book,
-- it will be added to newBooks table that is used for /lbreport
local function UpdateLorebooksData()
   for zone, zoneData in pairs(svData.lorebooks) do
      if zoneData ~= nil then
         for subzone, subzoneData in pairs(zoneData) do
            if subzoneData ~= nil then
               local epsilon = GetEpsilon(zone, subzone)
               for _, pinData in ipairs(subzoneData) do
                  if not InvalidPoint(pinData[1], pinData[2]) then
                     local knownBooks = LoreBooks_GetLocalData(zone, subzone)
                     if UnknownCoords(knownBooks, pinData[1], pinData[2], epsilon) then
                        SetLocalData(newBooks, zone, subzone, pinData)
                        LoreBooks_SetLocalData(zone, subzone, pinData)
                     end
                  end
               end
            end
         end
      end
   end
end

local function OnLoad(eventCode, name)
   if name ~= "LoreBooks" then return end
   EVENT_MANAGER:UnregisterForEvent("LoreBooks", EVENT_ADD_ON_LOADED)

   --saved variables for settings
   savedVariables = ZO_SavedVars:New("LBooks_SavedVariables", 2, nil, defaults)

   local pinTextureLevel = savedVariables.pinTexture.level
   local pinTextureSize = savedVariables.pinTexture.size
   local compassMaxDistance = savedVariables.compassMaxDistance

   local mapPinLayout_unknown = { level = pinTextureLevel, texture = GetPinTexture, size = pinTextureSize }
   local mapPinLayout_collected = { level = pinTextureLevel, texture = GetPinTexture, size = pinTextureSize, grayscale = IsPinGrayscale }
   local compassPinLayout = { maxDistance = compassMaxDistance, texture = pinTextures[savedVariables.pinTexture.type][2],
      sizeCallback = function(pin, angle, normalizedAngle, normalizedDistance)
         if zo_abs(normalizedAngle) > 0.25 then
            pin:SetDimensions(54 - 24 * zo_abs(normalizedAngle), 54 - 24 * zo_abs(normalizedAngle))
         else
            pin:SetDimensions(48, 48)
         end
      end,
      additionalLayout = {
         function(pin, angle, normalizedAngle, normalizedDistance)
            if (pinTexturesList[savedVariables.pinTexture.type] == "Shalidor's Library icons") then  --replace icon with icon from LoreLibrary
               local _, texture = GetLoreBookInfo(1, pin.pinTag[3], pin.pinTag[4])
               local icon = pin:GetNamedChild("Background")
               icon:SetTexture(texture)
            end
         end,
         function(pin)
            --I do not need to reset anything (texture is changed automatically), so the function is empty
         end
      }
   }

   --global saved variables for collecting lorebook loactions
   svData = ZO_SavedVars:NewAccountWide("LBooksData_SavedVariables", 4, nil, dataDefaults)

   --check if you have saved books that are not in default data
   UpdateLorebooksData()

   --initialize map pins
   LMP:AddPinType(PINS_UNKNOWN, MapCallback_unknown, nil, mapPinLayout_unknown, pinTooltipCreator)
   LMP:AddPinType(PINS_COLLECTED, MapCallback_collected, nil, mapPinLayout_collected, pinTooltipCreator)

   --add map filters
   LMP:AddPinFilter(PINS_UNKNOWN, GetString(LBOOKS_FILTER_UNKNOWN), nil, savedVariables.filters)
   LMP:AddPinFilter(PINS_COLLECTED, GetString(LBOOKS_FILTER_COLLECTED), nil, savedVariables.filters)

   --add handler for the left click
   LMP:SetClickHandlers(PINS_UNKNOWN, {
      [1] = {
         name = function(pin) return zo_strformat(LBOOKS_SET_WAYPOINT, GetLoreBookInfo(1, pin.m_PinTag[3], pin.m_PinTag[4])) end,
         show = function(pin) return not select(3, GetLoreBookInfo(1, pin.m_PinTag[3], pin.m_PinTag[4])) end,
         duplicates = function(pin1, pin2) return (pin1.m_PinTag[3] == pin2.m_PinTag[3] and pin1.m_PinTag[4] == pin2.m_PinTag[4]) end,
         callback = function(pin) PingMap(MAP_PIN_TYPE_PLAYER_WAYPOINT, MAP_TYPE_LOCATION_CENTERED, pin.normalizedX, pin.normalizedY) end,
      }
   })

   --initialize compass pins
   COMPASS_PINS:AddCustomPin(PINS_COMPASS, CompassCallback, compassPinLayout)
   COMPASS_PINS:RefreshPins(PINS_COMPASS)

   -- addon menu
   CreateSettingsMenu()

   --events
   EVENT_MANAGER:RegisterForEvent("LoreBooks", EVENT_LORE_BOOK_LEARNED, OnBookLearned)
end

EVENT_MANAGER:RegisterForEvent("LoreBooks", EVENT_ADD_ON_LOADED, OnLoad)
